-----------------------------------------------------------
--|> NEARBY LISTS
--: Manages the lists of players nearby
-----------------------------------------------------------
--|> Upvalue Globals
-----------------------------------------------------------
local _, NS = ...
local CopyTable = CopyTable
local pairs, tinsert, sort, wipe = pairs, tinsert, sort, wipe
local SM = LibStub("LibSharedMedia-3.0")

--|> GLOBALS
-----------------------------------------------------------
NS.CurrentNameplates = {}
NS.CurrentNameplatesSize = {}

NS.Nearby = {}

NS.NearbyListSize = 0
NS.NearbyList = {}

NS.ActiveList = {}
NS.InactiveList = {}
NS.ActiveDeadList = {}
NS.InactiveDeadList = {}

NS.PlayersOnBars = {}
NS.PlayersOnBarsSize = {}

--> Manage List Timeouts
local ActiveListCount = 0
function NS.ManageNearbyListTimeouts()
    local expired = false
    local removed = false
    local expiredCount = 0
    local count = 1
    local timestamp = GetTime()
    --: ACTIVE
    for player in pairs(NS.ActiveList) do
        if
            (timestamp - NS.ActiveList[player].TimeUpdated) > NS.Options.Sorting.NearbyActiveTimeout and
                NS.CurrentNameplates[player] == nil
         then
            NS.InactiveList[player] = NS.ActiveList[player]
            NS.InactiveList[player].TimeAdded = timestamp + (count * 0.001)
            NS.ActiveList[player] = nil
            count = count + 1
            expired = true
        end
    end
    ActiveListCount = count
    count = 0
    --: ACTIVE DEAD
    timestamp = GetTime()
    for player in pairs(NS.ActiveDeadList) do
        if (timestamp - NS.ActiveDeadList[player].TimeUpdated) > NS.Options.Sorting.NearbyActiveTimeout then
            NS.InactiveDeadList[player] = NS.ActiveDeadList[player]
            NS.InactiveDeadList[player].TimeAdded = timestamp + (count * 0.001)
            NS.ActiveDeadList[player] = nil
            expired = true
            count = count + 1
        end
    end
    --: INACTIVE
    timestamp = GetTime()
    for player in pairs(NS.InactiveList) do
        if (timestamp - NS.InactiveList[player].TimeUpdated) > NS.Options.Sorting.NearbyInactiveTimeout then
            NS.InactiveList[player] = nil
            NS.NearbyList[player] = nil
            NS.PlayerActiveCache[player] = nil
            expiredCount = expiredCount + 1
            expired = true
            removed = true
        end
    end
    --: INACTIVE DEAD
    timestamp = GetTime()
    for player in pairs(NS.InactiveDeadList) do
        if (timestamp - NS.InactiveDeadList[player].TimeUpdated) > NS.Options.Sorting.NearbyInactiveTimeout then
            NS.InactiveDeadList[player] = nil
            NS.NearbyList[player] = nil
            NS.PlayerActiveCache[player] = nil
            expiredCount = expiredCount + 1
            expired = true
            removed = true
        end
    end
    if expired or removed then
        NS.NearbyListSize = NS.NearbyListSize - expiredCount
        NS.SortNearbyList()
        NS.UpdateNearbyCount()
        NS.RefreshCurrentList()
        NS.CoreUI.ChangeTargetIcon()
    end
    if removed then
        NS.ManageBarsDisplayed()
    end
end

--> Update Player List
local function UpdatePlayerLists(GUID, timeUpdate, dead, newPlayerOnList)
    if not NS.Options.Bars then
        return
    end

    local reSortList = false
    local playerOnBar = true
    if NS.NearbyListSize <= NS.Options.Bars.MaxNumBars then
        if dead or newPlayerOnList then
            reSortList = true
        end
    else
        if NS.PlayersOnBars[GUID] ~= nil then
            if dead then
                reSortList = true
            end
        end
    end

    -- NEW PLAYER
    if NS.NearbyList[GUID] == nil or newPlayerOnList then -- ADDING NEW PLAYER
        playerOnBar = false
        reSortList = true
        --: New Player on KOS
        if NS.KosList[NS.PlayerActiveCache[GUID].Name] then
            -- Audio Alert
            if NS.Options.KOS.AudioAlert then
                NS.PlaySoundAlert(SM:Fetch("sound", NS.Options.KOS.AudioAlertFile), true)
            end
            -- Flash OS program icon (taskbar)
            if NS.Options.KOS.TaskbarAlert then
                FlashClientIcon()
            end
            -- Chat Alert
            if NS.Options.KOS.ChatAlert then
                NS.PrintAddonMessage(
                    "|TInterface/Addons/weizPVP/Media/Icons/kos.tga::0|t " ..
                        NS.FormatPlayerNameAndRealm(NS.PlayerActiveCache[GUID].Name) .. " |cff8fdaffdetected!|r "
                )
            end
        end
        --: New player
        if NS.Options.AudioAlerts.DetectedPlayerSound then -- NON-KOS PLAYER
            if not (NS.Options.AudioAlerts.DetectedPlayerSoundBGDisabled and NS.Zone.instance == "pvp") then
                if NS.Options.AudioAlerts.DisableInSanctuary == false and NS.Zone.pvpType ~= "sanctuary" then
                    NS.PlaySoundAlert(SM:Fetch("sound", NS.Options.AudioAlerts.DetectedPlayerSoundFile))
                end
            end
        end
        NS.NearbyList[GUID] = {}
        NS.NearbyList[GUID].TimeUpdated = timeUpdate
        NS.NearbyList[GUID].TimeAdded = timeUpdate
        if dead then
            NS.ActiveDeadList[GUID] = NS.ActiveDeadList[GUID] or {}
            NS.ActiveDeadList[GUID].TimeUpdated = timeUpdate
            NS.ActiveDeadList[GUID].TimeAdded = timeUpdate
        else
            NS.ActiveList[GUID] = NS.ActiveList[GUID] or {}
            NS.ActiveList[GUID].TimeAdded = timeUpdate
            NS.ActiveList[GUID].TimeUpdated = timeUpdate
        end
    elseif not NS.ActiveList[GUID] and not NS.ActiveDeadList[GUID] then -- EXISTING PLAYER; WAS INACTIVE
        reSortList = true
        if dead then
            NS.ActiveDeadList[GUID] = NS.InactiveDeadList[GUID] or {}
            NS.ActiveDeadList[GUID].TimeUpdated = timeUpdate
            NS.ActiveDeadList[GUID].TimeAdded = timeUpdate
            NS.NearbyList[GUID].TimeUpdated = timeUpdate
            NS.NearbyList[GUID].TimeAdded = timeUpdate
            NS.ActiveList[GUID] = nil
        else
            NS.ActiveList[GUID] = NS.InactiveList[GUID] or {}
            NS.ActiveList[GUID].TimeUpdated = timeUpdate
            NS.ActiveList[GUID].TimeAdded = timeUpdate
            NS.NearbyList[GUID].TimeUpdated = timeUpdate
            NS.NearbyList[GUID].TimeAdded = timeUpdate
            NS.ActiveDeadList[GUID] = nil
        end
        NS.InactiveList[GUID] = nil
        NS.InactiveDeadList[GUID] = nil
    else -- EXISTING PLAYER; ACTIVE
        if dead then
            NS.ActiveDeadList[GUID] = NS.ActiveDeadList[GUID] or {}
            NS.ActiveDeadList[GUID].TimeUpdated = timeUpdate
            NS.NearbyList[GUID].TimeUpdated = timeUpdate
            NS.ActiveList[GUID] = nil
        else
            NS.ActiveList[GUID] = NS.ActiveList[GUID] or {}
            NS.ActiveList[GUID].TimeUpdated = timeUpdate
            NS.NearbyList[GUID].TimeUpdated = timeUpdate
            NS.ActiveDeadList[GUID] = nil
        end
    end
    --> Check to see if we need to add a kos player to the bars
    if not playerOnBar and ActiveListCount > NS.Options.Bars.MaxNumBars then
        if not NS.KosList[NS.PlayerActiveCache[GUID].Name] then
            return
        else
            reSortList = true
            newPlayerOnList = true
        end
    end
    -- Sort only if we moved the player from one sub-list to another
    if reSortList then
        NS.SortNearbyList()
        if NS.PlayersOnBars[GUID] or newPlayerOnList then
            NS.RefreshCurrentList()
        end
    else
        if NS.PlayersOnBars[GUID] then
            NS.RefreshBarByGUID(GUID)
        end
    end
end

--> Add Player Data
local lastTimestamp = GetTime()
function NS.AddPlayerDataToNearbyList(GUID, newPlayerOnList)
    if not GUID then
        return
    end
    local dead = false
    if NS.PlayerActiveCache[GUID] and NS.PlayerActiveCache[GUID].Dead ~= nil then
        dead = NS.PlayerActiveCache[GUID].Dead
    else
        NS.PlayerActiveCache[GUID].Dead = false
    end
    local time = GetTime()
    if lastTimestamp >= time then
        lastTimestamp = lastTimestamp + 0.0001
    else
        lastTimestamp = time
    end
    if newPlayerOnList then
        NS.NearbyListSize = NS.NearbyListSize + 1
        NS.UpdateNearbyCount()
    end
    UpdatePlayerLists(GUID, lastTimestamp, dead, newPlayerOnList)
end

--> Refresh Current List
function NS.RefreshNearbyList()
    -- Refreshing all bars due to required sorting
    local localPlayersOnBars = {}
    for k, data in pairs(NS.CurrentList) do
        if k <= NS.Options.Bars.MaxNumBars then
            -- Update Alpha if needed
            local alpha = NS.Options.Bars.AlphaDefault or 1
            if NS.InactiveList[data.player] or NS.InactiveDeadList[data.player] then
                alpha = NS.Options.Bars.AlphaInactive
            elseif NS.ActiveDeadList[data.player] then
                alpha = NS.Options.Bars.AlphaDead
            end
            if NS.PlayerActiveCache[data.player] then
                localPlayersOnBars[NS.PlayerActiveCache[data.player].GUID] = k
                local playerName = NS.PlayerActiveCache[data.player].Name or nil
            -- Update Bar
            -- UpdateBar(
            --   k,
            --   data.player,
            --   alpha,
            --   NS.PlayerActiveCache[data.player].Health,
            --   NS.PlayerDB[playerName].C,
            --   NS.PlayerDB[playerName].G,
            --   NS.PlayerDB[playerName].L,
            --   NS.PlayerDB[playerName].E,
            --   NS.PlayerActiveCache[data.player].Stealth,
            --   NS.PlayerActiveCache[data.player].Dead,
            --   NS.PlayerDB[playerName].RL,
            --   playerName
            -- )
            end
        else
            break
        end
    end
    NS.PlayersOnBars = CopyTable(localPlayersOnBars)
end

--> Sort Nearby List
local tempCurrentList = {}
local tempActiveList = {}
local tempActiveDeadList = {}
local tempInactiveList = {}
local tempInactiveDeadList = {}
function NS.SortNearbyList()
    --: ActiveList
    for player in pairs(NS.ActiveList) do
        if NS.NearbyList[player] then
            if NS.NearbyList[player].TimeAdded then
                tinsert(tempActiveList, {player = player, time = NS.NearbyList[player].TimeAdded})
            end
        end
    end
    --: ActiveDeadList
    for player in pairs(NS.ActiveDeadList) do
        if NS.NearbyList[player] then
            if NS.NearbyList[player].TimeAdded then
                tinsert(tempActiveDeadList, {player = player, time = NS.NearbyList[player].TimeAdded})
            end
        end
    end
    --: InactiveDeadList
    for player in pairs(NS.InactiveDeadList) do
        if NS.NearbyList[player] then
            if NS.NearbyList[player].TimeAdded then
                tinsert(tempInactiveDeadList, {player = player, time = NS.NearbyList[player].TimeAdded})
            end
        end
    end
    --: InactiveList
    for player in pairs(NS.InactiveList) do
        if NS.NearbyList[player] then
            if NS.NearbyList[player].TimeAdded then
                tinsert(tempInactiveList, {player = player, time = NS.NearbyList[player].TimeAdded})
            end
        end
    end
    --: sorts
    sort(
        tempActiveList,
        function(a, b)
            return a.time < b.time
        end
    )
    sort(
        tempActiveDeadList,
        function(a, b)
            return a.time < b.time
        end
    )
    sort(
        tempInactiveList,
        function(a, b)
            return a.time < b.time
        end
    )
    sort(
        tempInactiveDeadList,
        function(a, b)
            return a.time < b.time
        end
    )
    --: create player list
    for player in pairs(tempActiveList) do
        if NS.PlayerActiveCache[tempActiveList[player].player] then
            if NS.KosList[NS.PlayerActiveCache[tempActiveList[player].player].Name] then
                tinsert(tempCurrentList, tempActiveList[player])
                tempActiveList[player] = nil
            end
        end
    end
    for player in pairs(tempActiveList) do
        tinsert(tempCurrentList, tempActiveList[player])
    end
    for player in pairs(tempActiveDeadList) do
        tinsert(tempCurrentList, tempActiveDeadList[player])
    end
    for player in pairs(tempInactiveList) do
        tinsert(tempCurrentList, tempInactiveList[player])
    end
    for player in pairs(tempInactiveDeadList) do
        tinsert(tempCurrentList, tempInactiveDeadList[player])
    end
    NS.CurrentList = CopyTable(tempCurrentList)
    wipe(tempCurrentList)
    wipe(tempActiveList)
    wipe(tempActiveDeadList)
    wipe(tempInactiveList)
    wipe(tempInactiveDeadList)
end

--> Clear List Data
function NS.ClearNearbyListData()
    if NS.NearbyCount ~= 0 then
        wipe(NS.CurrentList)
        wipe(NS.CurrentNameplates)
        wipe(NS.NearbyList)
        wipe(NS.ActiveList)
        wipe(NS.InactiveList)
        wipe(NS.ActiveDeadList)
        wipe(NS.InactiveDeadList)
        wipe(NS.PlayersOnBars)
        wipe(NS.PlayerActiveCache)
        NS.NearbyListSize = 0
        NS.SortNearbyList()
        NS.ManageBarsDisplayed()
        NS.RefreshCurrentList()
        NS.UpdateNearbyCount()
        weizPVP_CoreFrame.ScrollFrame:SetVerticalScroll(0)
        NS:PlayerTargetEvent()
        NS.AutoResize()
        NS.CoreUI.ChangeTargetIcon()
    end
end
